/*
 * Licensed Materials - Property of IBM
 *
 * com.ibm.etools.c
 *
 *  Copyright IBM Corporation 2004. All Rights Reserved. 
 *
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.etools.c.importer.AST;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.eclipse.cdt.core.parser.ast.ASTClassKind;
import org.eclipse.cdt.core.parser.ast.IASTAbstractDeclaration;
import org.eclipse.cdt.core.parser.ast.IASTAbstractTypeSpecifierDeclaration;
import org.eclipse.cdt.core.parser.ast.IASTClassSpecifier;
import org.eclipse.cdt.core.parser.ast.IASTCompilationUnit;
import org.eclipse.cdt.core.parser.ast.IASTExpression;
import org.eclipse.cdt.core.parser.ast.IASTField;
import org.eclipse.cdt.core.parser.ast.IASTInclusion;
import org.eclipse.cdt.core.parser.ast.IASTTypeSpecifier;
import org.eclipse.cdt.core.parser.ast.IASTVariable;
import org.eclipse.cdt.internal.core.parser.ast.complete.ASTAbstractTypeSpecifierDeclaration;
import org.eclipse.cdt.internal.core.parser.ast.complete.ASTClassSpecifier;
import org.eclipse.cdt.internal.core.parser.ast.complete.ASTElaboratedTypeSpecifier;
import org.eclipse.cdt.internal.core.parser.ast.complete.ASTEnumerationSpecifier;
import org.eclipse.cdt.internal.core.parser.ast.complete.ASTEnumerator;
import org.eclipse.cdt.internal.core.parser.ast.complete.ASTSimpleTypeSpecifier;
import org.eclipse.cdt.internal.core.parser.ast.complete.ASTTypedef;
import org.eclipse.cdt.internal.core.parser.ast.complete.ASTVariable;

import com.ibm.etools.c.CArray;
import com.ibm.etools.c.CClassifier;
import com.ibm.etools.c.CDerivableType;
import com.ibm.etools.c.CEnumerator;
import com.ibm.etools.c.CFactory;
import com.ibm.etools.c.CField;
import com.ibm.etools.c.CParameter;
import com.ibm.etools.c.CStruct;
import com.ibm.etools.c.CStructured;
import com.ibm.etools.c.CTypedef;
import com.ibm.etools.c.CUnion;
import com.ibm.etools.c.datatypes.CDirectionKind;
import com.ibm.etools.c.datatypes.CEnumeration;
import com.ibm.etools.c.datatypes.DatatypesFactory;
import com.ibm.etools.c.importer.CModelConverter;
import com.ibm.etools.c.importer.CPreferenceStore;
import com.ibm.etools.c.importer.CQuery;
import com.ibm.etools.c.importer.CResource;
import com.ibm.etools.c.importer.CSourceLocation;
import com.ibm.etools.c.importer.CTypeDescriptorBuilder;
import com.ibm.etools.c.importer.CTypeTable;
import com.ibm.etools.c.importer.CTypedElementParser;
import com.ibm.etools.typedescriptor.InstanceTDBase;

/**
 * 
 * @author Bernhard Jakob, IBM
 * 
 * CSourceUnit represents a source unit which can be a C source file or header
 * file. It has an associated object list which contains all of the type and
 * function objects within this source file. The importSource() method is the
 * exported interface.
 *  
 */
public class CSourceUnit {

    public static final String copyright = "Licensed Material - Property of IBM\n com.ibm.etools.c\n(C) Copyright IBM Corp. 2004 - All Rights Reserved.\nUS Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp."; //$NON-NLS-1$

    private CModelConverter _modelConverter;

    // The file name of the source unit
    private String _fileName;

    // parsed source
    private CParseCallback _sourceParseCallback;

    // The list to keep all the C model objects in this source unit
    private ArrayList _objectList;

    // The list to keep all top level types.
    private ArrayList _topTypesList;

    // The type table from CModelConverter
    private CTypeTable _typeTable;

    // The type parser from CModelConverter
    private CTypeParser _typeParser;

    // The list of all include files
    private ArrayList _includeFiles;

    // A factory used to generate C language entities.
    private CFactory _factory;

    // A factory used to generate C datatypes.
    private DatatypesFactory _datatypesFactory;

    // Is this the first pass thru the file
    private boolean _firstPass;

    private int structureCounter = 1;

    public static final String STRUCTURE_TYPE_SPECIFIER = "_noNameStructure";

    private String currentStructureName = "";

    private int unionCounter = 1;

    public static final String UNION_TYPE_SPECIFIER = "_noNameUnion";

    private String currentUnionName = "";

    private int enumCounter = 1;

    public static final String ENUM_TYPE_SPECIFIER = "_noNameEnum";

    private String currentEnumName = "";

    private boolean useFieldStack = false;

    private Vector addFieldStack = new Vector(1);

    /**
     * 
     * constructor
     * 
     * Create a CSourceUnit object from a given source file name and a
     * CModelConverter object.
     *  
     */
    public CSourceUnit(String fileName, CModelConverter modelConverter) {
        _fileName = fileName;
        _modelConverter = modelConverter;
        _sourceParseCallback = null;
        _objectList = new ArrayList();
        _topTypesList = new ArrayList();
        _typeTable = modelConverter.getTypeTable();
        _typeParser = modelConverter.getASTTypeParser();
        _includeFiles = new ArrayList();
        _factory = CFactory.eINSTANCE;
        _datatypesFactory = DatatypesFactory.eINSTANCE;
        _firstPass = true;
    }

    /**
     * 
     * constructor
     * 
     * Create a CSourceUnit object from a given DataElement object which
     * represents the source file and a CModelConverter object.
     *  
     */
    public CSourceUnit(CParseCallback cParseCallback, CModelConverter modelConverter) {
        _sourceParseCallback = cParseCallback;
        _fileName = cParseCallback.getFileName();
        _modelConverter = modelConverter;
        _objectList = new ArrayList();
        _topTypesList = new ArrayList();
        _typeTable = modelConverter.getTypeTable();
        _typeParser = modelConverter.getASTTypeParser();
        _includeFiles = new ArrayList();
        _factory = CFactory.eINSTANCE;
        _datatypesFactory = DatatypesFactory.eINSTANCE;
        _firstPass = true;
    }

    /**
     * 
     * Add the parsed content of a single source file to the EMF object list
     * This method is the entry point of the parsing process.
     * 
     * @return java.util.ArrayList The list of C model objects created from the
     *         parsed information.
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    public ArrayList importSource() throws Exception {

        IASTCompilationUnit compilationUnit = null;

        /*
         * do some inititalization
         */
        _objectList.clear();
        _topTypesList.clear();

        //
        // Add all include files to the include list
        //
        Iterator inclusionsIterator = _sourceParseCallback.getInclusions();

        while (inclusionsIterator.hasNext()) {
            IASTInclusion iASTInclusion = (IASTInclusion) inclusionsIterator.next();
            String inclusionName = iASTInclusion.getName();

            _includeFiles.add(inclusionName);
        }

        //
        // get and examine/process parsed source
        //
        compilationUnit = _sourceParseCallback.getCompilationUnit();

        addNode(compilationUnit);

        _firstPass = false;

        /*
         * second pass is no more necessary
         */
        //addNode(compilationUnit);
        // Set the offset and size for all struct/union fields.
        // We do this as the last step when all the type information is
        // known.
        if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
            for (int i = 0; i < _objectList.size(); i++) {
                Object object = _objectList.get(i);
                if (object instanceof CStructured)
                    CTypeDescriptorBuilder.setInstancesOffsetAndSize((CStructured) object);
            }
        }

        currentEnumName = null;
        currentStructureName = null;
        currentUnionName = null;

        return _topTypesList;
    }

    /**
     * 
     * Add the parsed content of a single source file to the EMF object list
     * This method is the entry point of the parsing process.
     * 
     * @param IASTCompilationUnit
     *            contains the parsed source
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private void addNode(IASTCompilationUnit compilationUnit) throws Exception {

        //
        // get and examine/process parsed source
        //
        Iterator compilationUnitIterator = compilationUnit.getDeclarations();

        while (compilationUnitIterator.hasNext()) {
            Object compilationUnitObject = compilationUnitIterator.next();

            if (compilationUnitObject instanceof ASTAbstractTypeSpecifierDeclaration) {
                ASTAbstractTypeSpecifierDeclaration aSTAbstractTypeSpecifierDeclaration = (ASTAbstractTypeSpecifierDeclaration) compilationUnitObject;
                Object aSTTypeSpecifier = aSTAbstractTypeSpecifierDeclaration.getTypeSpecifier();

                addNode(aSTTypeSpecifier, true);

            } else if (compilationUnitObject instanceof ASTClassSpecifier) {
                addNode((ASTClassSpecifier) compilationUnitObject, true);

            } else if (compilationUnitObject instanceof ASTTypedef) {
                addNode((ASTTypedef) compilationUnitObject, true);

            } else if (compilationUnitObject instanceof ASTEnumerationSpecifier) {
                addNode((ASTEnumerationSpecifier) compilationUnitObject, true);

            } else if (compilationUnitObject instanceof ASTVariable) {
                addNode((ASTVariable) compilationUnitObject, true);
            }
        }
    }

    /**
     * 
     * Add a node of the compilationUnit to the EMF object list
     * 
     * @param Object
     *            a node of the compilationUnit to be added to the EMF object
     *            list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private void addNode(Object object, boolean topLevel) throws Exception {

        if (object instanceof ASTClassSpecifier) {
            addNode((ASTClassSpecifier) object, topLevel);

        } else if (object instanceof ASTEnumerationSpecifier) {
            addNode((ASTEnumerationSpecifier) object, topLevel);

        } else if (object instanceof ASTTypedef) {
            addNode((ASTTypedef) object, topLevel);

        } else if (object instanceof ASTElaboratedTypeSpecifier) {
            addNode((ASTElaboratedTypeSpecifier) object, topLevel);
        }
    }

    /**
     * 
     * Add a node of the compilationUnit to the EMF object list
     * 
     * @param IASTClassSpecifier
     *            a node of the compilationUnit to be added to the EMF object
     *            list
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private void addNode(IASTClassSpecifier aSTClassSpecifier, boolean topLevel) throws Exception {

        if (aSTClassSpecifier != null) {
            ASTClassKind aSTClassKind = aSTClassSpecifier.getClassKind();

            if (aSTClassKind == ASTClassKind.STRUCT) {
                addStruct(aSTClassSpecifier, topLevel);

            } else if (aSTClassKind == ASTClassKind.UNION) {
                addUnion(aSTClassSpecifier, topLevel);

            } else if (aSTClassKind == ASTClassKind.ENUM) {
                addEnum(aSTClassSpecifier, topLevel);
            }

            // Ignore other tokens
        }
    }

    /**
     * 
     * Add a node of the compilationUnit to the EMF object list
     * 
     * @param IASTVariable
     *            a node of the compilationUnit to be added to the EMF object
     *            list
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private void addNode(IASTVariable aSTVariable, boolean topLevel) throws Exception {

        IASTAbstractDeclaration aSTAbstractDeclaration = aSTVariable.getAbstractDeclaration();
        Object object = aSTAbstractDeclaration.getTypeSpecifier();

        addNode(object, topLevel);

    }

    /**
     * 
     * Add a node of the compilationUnit to the EMF object list
     * 
     * @param ASTTypedef
     *            a typedef node of the compilationUnit to be added to the EMF
     *            object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private void addNode(ASTTypedef aSTTypedef, boolean topLevel) throws Exception {

        /*
         * first add concrete type ...
         */
        IASTAbstractDeclaration iASTAbstractDeclaration = aSTTypedef.getAbstractDeclarator();
        Object fieldObject = iASTAbstractDeclaration.getTypeSpecifier();

        if (fieldObject instanceof ASTClassSpecifier || fieldObject instanceof ASTEnumerationSpecifier || fieldObject instanceof ASTElaboratedTypeSpecifier) {
            addNode(fieldObject, topLevel);
        }

        /*
         * ... then add type definition
         */
        addTypedef(aSTTypedef, topLevel);
    }

    /**
     * 
     * Add a node of the compilationUnit to the EMF object list
     * 
     * @param ASTEnumerationSpecifier
     *            a enumeration node of the compilationUnit to be added to the
     *            EMF object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private void addNode(ASTEnumerationSpecifier aSTEnumerationSpecifier, boolean topLevel) throws Exception {

        addEnum(aSTEnumerationSpecifier, topLevel);
    }

    /**
     * 
     * Add a node of the compilationUnit to the EMF object list
     * 
     * @param ASTEnumerationSpecifier
     *            a enumeration node of the compilationUnit to be added to the
     *            EMF object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private void addNode(ASTElaboratedTypeSpecifier aSTElaboratedTypeSpecifier, boolean topLevel) throws Exception {

        ASTClassKind aSTClassKind = aSTElaboratedTypeSpecifier.getClassKind();

        List list = aSTElaboratedTypeSpecifier.getReferences();

        if (aSTClassKind == ASTClassKind.STRUCT) {
            // nothing to do

        } else if (aSTClassKind == ASTClassKind.UNION) {
            // nothing to do

        } else if (aSTClassKind == ASTClassKind.ENUM) {
            // nothing to do
        }
    }

    /**
     * 
     * Add a node representing a C structure to the object list
     * 
     * @param IASTClassSpecifier
     *            a structure node of the compilationUnit to be added to the EMF
     *            object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CStruct addStruct(IASTClassSpecifier structToken, boolean topLevel) throws Exception {

        String structName = getNameofStructure(structToken);

        String oldStructureName = new String(currentStructureName);

        currentStructureName = new String(structName);

        if (useFieldStack) {
            addFieldStack.add(new String(structName));
        }

        CStruct cStruct = generateCStruct(structName);

        if (cStruct != null && cStruct.getContains().size() == 0) {

            initializeCStruct(cStruct, structToken, topLevel);

            if (topLevel) {
                _topTypesList.add(cStruct);
            }

            _objectList.add(cStruct);
        }

        currentStructureName = new String(oldStructureName);

        return cStruct;
    }

    /**
     * 
     * creates and returnes a CStruct
     * 
     * @param String
     *            name of structure
     *  
     */
    private CStruct generateCStruct(String structName) {

        CStruct cStruct = (CStruct) _typeTable.getType("struct " + structName);

        if (cStruct == null) {
            cStruct = _factory.createCStruct();

            cStruct.setName(structName);

            _typeTable.addType("struct " + structName, cStruct);
        }

        return cStruct;
    }

    /**
     * 
     * Add a node representing a C structure to the object list
     * 
     * @param IASTClassSpecifier
     *            a structure node of the compilationUnit to be added to the EMF
     *            object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CStruct initializeCStruct(CStruct cStruct, IASTClassSpecifier structToken, boolean topLevel) throws Exception {

        _typeTable.addSourceLocation(cStruct, new CSourceLocation(_fileName));

        Iterator structIterator = structToken.getDeclarations();

        //
        // loop over all elements of passed structure
        //
        while (structIterator.hasNext()) {
            Object structObject = structIterator.next();

            if (structObject instanceof IASTField) {
                cStruct = initializeCStructWithField(cStruct, (IASTField) structObject, topLevel);

            } else if (structObject instanceof ASTEnumerationSpecifier) {
                cStruct = initializeCStructWithEnum(cStruct, (ASTEnumerationSpecifier) structObject, topLevel);

            } else if (structObject instanceof IASTClassSpecifier) {
                IASTClassSpecifier aSTClassSpecifier = (IASTClassSpecifier) structObject;
                ASTClassKind aSTClassKind = aSTClassSpecifier.getClassKind();

                if (aSTClassKind == ASTClassKind.STRUCT) {
                    cStruct = initializeCStructWithStructure(cStruct, aSTClassSpecifier, topLevel);

                } else if (aSTClassKind == ASTClassKind.UNION) {
                    cStruct = initializeCStructWithUnion(cStruct, aSTClassSpecifier, topLevel);

                } else if (aSTClassKind == ASTClassKind.ENUM) {
                    cStruct = initializeCStructWithEnum(cStruct, aSTClassSpecifier, topLevel);
                }
            } else if (structObject instanceof IASTAbstractTypeSpecifierDeclaration) {
                IASTAbstractTypeSpecifierDeclaration aSTAbstractTypeSpecifierDeclaration = (IASTAbstractTypeSpecifierDeclaration) structObject;
                IASTTypeSpecifier aSTTypeSpecifier = aSTAbstractTypeSpecifierDeclaration.getTypeSpecifier();

                if (aSTTypeSpecifier instanceof IASTClassSpecifier) {
                    IASTClassSpecifier aSTClassSpecifier = (IASTClassSpecifier) aSTTypeSpecifier;
                    ASTClassKind aSTClassKind = aSTClassSpecifier.getClassKind();

                    if (aSTClassKind == ASTClassKind.STRUCT) {
                        cStruct = initializeCStructWithStructure(cStruct, aSTClassSpecifier, topLevel);

                    } else if (aSTClassKind == ASTClassKind.UNION) {
                        cStruct = initializeCStructWithUnion(cStruct, aSTClassSpecifier, topLevel);

                    } else if (aSTClassKind == ASTClassKind.ENUM) {
                        cStruct = initializeCStructWithEnum(cStruct, aSTClassSpecifier, topLevel);
                    }
                }
            }
        }

        return cStruct;
    }

    /**
     * 
     * Add a node representing a C structure to the object list
     * 
     * @param IASTField
     *            a AST field of the compilationUnit to be added to the passed
     *            structure
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CStruct initializeCStructWithField(CStruct cStruct, IASTField aSTField, boolean topLevel) throws Exception {

        IASTAbstractDeclaration iASTAbstractDeclaration = aSTField.getAbstractDeclaration();
        Object fieldObject = iASTAbstractDeclaration.getTypeSpecifier();

        if (fieldObject instanceof ASTSimpleTypeSpecifier || fieldObject instanceof ASTClassSpecifier || fieldObject instanceof ASTEnumerationSpecifier
                || fieldObject instanceof ASTElaboratedTypeSpecifier) {

            CField afield = addField(aSTField);
            cStruct.getContains().add(afield);

            if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
                afield.setInstanceTDBase(CTypeDescriptorBuilder.createInstanceForField(afield));
            }
        }

        return cStruct;
    }

    /**
     * 
     * Add a node representing a C structure to the object list
     * 
     * @param IASTField
     *            a AST field of the compilationUnit to be added to the passed
     *            structure
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CStruct initializeCStructWithStructure(CStruct cStruct, IASTClassSpecifier aSTClassSpecifier, boolean topLevel) throws Exception {

        // Nested types also live in the global namespace in C.
        // This will add the nested type to the type table and
        // the object list.
        CStruct nestedStruct = addStruct(aSTClassSpecifier, false);

        // If the nested struct contributes to the size of the
        // outer
        // struct, we will create a clone of it and add the
        // cloned
        // struct to the outer struct's containing list.
        // The cloned struct has its own fields. The Type
        // Descriptor
        // instances of the fields may have different offsets
        // than
        // the original struct.
        if (_typeParser.nestedTypeHasSizeEffect(aSTClassSpecifier)) {
            CStruct clonedStruct = clonedStruct(nestedStruct);
            cStruct.getContains().add(clonedStruct);
        }

        return cStruct;
    }

    /**
     * 
     * Add a node representing a C structure to the object list
     * 
     * @param IASTField
     *            a AST field of the compilationUnit to be added to the passed
     *            structure
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CStruct initializeCStructWithUnion(CStruct cStruct, IASTClassSpecifier aSTClassSpecifier, boolean topLevel) throws Exception {

        CUnion nestedUnion = addUnion(aSTClassSpecifier, false);

        if (_typeParser.nestedTypeHasSizeEffect(aSTClassSpecifier)) {
            CUnion clonedUnion = clonedUnion(nestedUnion);
            cStruct.getContains().add(clonedUnion);
        }

        return cStruct;
    }

    /**
     * 
     * Add a node representing a C structure to the object list
     * 
     * @param IASTField
     *            a AST field of the compilationUnit to be added to the passed
     *            structure
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CStruct initializeCStructWithEnum(CStruct cStruct, IASTClassSpecifier aSTClassSpecifier, boolean topLevel) throws Exception {

        CEnumeration nestedEnum = addEnum(aSTClassSpecifier, true);

        return cStruct;
    }

    /**
     * 
     * Add a node representing a C structure to the object list
     * 
     * @param IASTField
     *            a AST field of the compilationUnit to be added to the passed
     *            structure
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CStruct initializeCStructWithEnum(CStruct cStruct, ASTEnumerationSpecifier aSTEnumerationSpecifier, boolean topLevel) throws Exception {

        CEnumeration nestedEnum = addEnum(aSTEnumerationSpecifier, topLevel);

        return cStruct;
    }

    /**
     * 
     * Add a node representing a C union to the object list
     * 
     * @param IASTClassSpecifier
     *            an union node of the compilationUnit to be added to the EMF
     *            object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CUnion addUnion(IASTClassSpecifier unionToken, boolean topLevel) throws Exception {

        String unionName = getNameofUnion(unionToken);

        String oldUnionName = new String(currentUnionName);

        currentUnionName = new String(unionName);

        if (useFieldStack) {
            addFieldStack.add(new String(unionName));
        }

        CUnion cUnion = generateCUnion(unionName);

        if (cUnion != null && cUnion.getContains().size() == 0) {

            initializeCUnion(cUnion, unionToken, topLevel);

            if (topLevel) {
                _topTypesList.add(cUnion);
            }

            _objectList.add(cUnion);
        }

        currentUnionName = new String(oldUnionName);

        return cUnion;
    }

    /**
     * 
     * creates and returns a CUnion
     * 
     * @param String
     *            name of union
     *  
     */
    private CUnion generateCUnion(String unionName) {

        CUnion cUnion = (CUnion) _typeTable.getType("union " + unionName);

        if (cUnion == null) {
            cUnion = _factory.createCUnion();

            cUnion.setName(unionName);

            _typeTable.addType("union " + unionName, cUnion);
        }

        return cUnion;
    }

    /**
     * 
     * Add a node representing a C structure to the object list
     * 
     * @param IASTClassSpecifier
     *            a structure node of the compilationUnit to be added to the EMF
     *            object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CUnion initializeCUnion(CUnion cUnion, IASTClassSpecifier unionToken, boolean topLevel) throws Exception {

        _typeTable.addSourceLocation(cUnion, new CSourceLocation(_fileName));

        Iterator unionIterator = unionToken.getDeclarations();

        //
        // loop over all elements of passed union
        //
        while (unionIterator.hasNext()) {
            Object unionObject = unionIterator.next();

            if (unionObject instanceof IASTField) {
                cUnion = initializeCUnionWithField(cUnion, (IASTField) unionObject, topLevel);

            } else if (unionObject instanceof ASTEnumerationSpecifier) {
                cUnion = initializeCUnionWithEnum(cUnion, (ASTEnumerationSpecifier) unionObject, topLevel);

            } else if (unionObject instanceof IASTClassSpecifier) {
                IASTClassSpecifier aSTClassSpecifier = (IASTClassSpecifier) unionObject;
                ASTClassKind aSTClassKind = aSTClassSpecifier.getClassKind();

                if (aSTClassKind == ASTClassKind.STRUCT) {
                    cUnion = initializeCUnionWithStructure(cUnion, aSTClassSpecifier, topLevel);

                } else if (aSTClassKind == ASTClassKind.UNION) {
                    cUnion = initializeCUnionWithUnion(cUnion, aSTClassSpecifier, topLevel);

                } else if (aSTClassKind == ASTClassKind.ENUM) {
                    cUnion = initializeCUnionWithEnum(cUnion, aSTClassSpecifier, topLevel);
                }
            }
        }

        return cUnion;
    }

    /**
     * 
     * Add a node representing a C union to the object list
     * 
     * @param IASTField
     *            a AST field of the compilationUnit to be added to the passed
     *            structure
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CUnion initializeCUnionWithField(CUnion cUnion, IASTField aSTField, boolean topLevel) throws Exception {

        IASTAbstractDeclaration aSTAbstractDeclaration = aSTField.getAbstractDeclaration();
        Object fieldObject = aSTAbstractDeclaration.getTypeSpecifier();

        if (fieldObject instanceof ASTSimpleTypeSpecifier || fieldObject instanceof ASTClassSpecifier || fieldObject instanceof ASTEnumerationSpecifier
                || fieldObject instanceof ASTElaboratedTypeSpecifier) {

            CField cField = addField(aSTField);
            cUnion.getContains().add(cField);

            if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
                cField.setInstanceTDBase(CTypeDescriptorBuilder.createInstanceForField(cField));
            }
        }

        return cUnion;
    }

    /**
     * 
     * Add a node representing a C union to the object list
     * 
     * @param IASTField
     *            a AST field of the compilationUnit to be added to the passed
     *            structure
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CUnion initializeCUnionWithStructure(CUnion cUnion, IASTClassSpecifier aSTClassSpecifier, boolean topLevel) throws Exception {

        // Nested types also live in the global namespace in C.
        // This will add the nested type to the type table and
        // the
        // object list.
        CStruct nestedStruct = addStruct(aSTClassSpecifier, false);

        // If the nested struct contributes to the size of the
        // outer
        // struct, we will create a clone of it and add the
        // cloned
        // struct to the outer struct's containing list.
        // The cloned struct has its own fields. The Type
        // Descriptor
        // instances of the fields may have different offsets
        // than the original struct.
        if (_typeParser.nestedTypeHasSizeEffect(aSTClassSpecifier)) {
            CStruct clonedStruct = clonedStruct(nestedStruct);
            cUnion.getContains().add(clonedStruct);
        }

        return cUnion;
    }

    /**
     * 
     * Add a node representing a C union to the object list
     * 
     * @param IASTField
     *            a AST field of the compilationUnit to be added to the passed
     *            structure
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CUnion initializeCUnionWithUnion(CUnion cUnion, IASTClassSpecifier aSTClassSpecifier, boolean topLevel) throws Exception {

        CUnion nestedUnion = addUnion(aSTClassSpecifier, false);

        if (_typeParser.nestedTypeHasSizeEffect(aSTClassSpecifier)) {
            CUnion clonedUnion = clonedUnion(nestedUnion);
            cUnion.getContains().add(clonedUnion);
        }

        return cUnion;
    }

    /**
     * 
     * Add a node representing a C union to the object list
     * 
     * @param IASTField
     *            a AST field of the compilationUnit to be added to the passed
     *            structure
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CUnion initializeCUnionWithEnum(CUnion cUnion, IASTClassSpecifier aSTClassSpecifier, boolean topLevel) throws Exception {

        CEnumeration nestedEnumeration = addEnum(aSTClassSpecifier, true);

        return cUnion;
    }

    /**
     * 
     * Add a node representing a C union to the object list
     * 
     * @param IASTField
     *            a AST field of the compilationUnit to be added to the passed
     *            structure
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CUnion initializeCUnionWithEnum(CUnion cUnion, ASTEnumerationSpecifier aSTEnumerationSpecifier, boolean topLevel) throws Exception {

        CEnumeration nestedEnumeration = addEnum(aSTEnumerationSpecifier, topLevel);

        return cUnion;
    }

    /**
     * 
     * Add an enum to the object list
     * 
     * @param IASTClassSpecifier
     *            an union node of the compilationUnit to be added to the EMF
     *            object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CEnumeration addEnum(IASTClassSpecifier enumToken, boolean topLevel) throws Exception {

        String enumerationName = enumToken.getName();

        CEnumeration cEnumeration = generateCEnumeration(enumerationName);

        if (cEnumeration != null && cEnumeration.getEnumerator().size() == 0) {

            cEnumeration = initializeCEnumeration(cEnumeration, enumToken, topLevel);

            _objectList.add(cEnumeration);
            _topTypesList.add(cEnumeration);
        }

        return cEnumeration;
    }

    /**
     * 
     * Add an enum to the object list
     * 
     * @param ASTEnumerationSpecifier
     *            an union node of the compilationUnit to be added to the EMF
     *            object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CEnumeration addEnum(ASTEnumerationSpecifier enumToken, boolean topLevel) throws Exception {

        String enumerationName = enumToken.getName();

        CEnumeration cEnumeration = generateCEnumeration(enumerationName);

        if (cEnumeration != null && cEnumeration.getEnumerator().size() == 0) {

            cEnumeration = initializeCEnumeration(cEnumeration, enumToken, topLevel);

            _objectList.add(cEnumeration);
            _topTypesList.add(cEnumeration);
        }

        return cEnumeration;
    }

    /**
     * 
     * creates and returns a CEnumeration
     * 
     * @param String
     *            name of enumeration
     *  
     */
    private CEnumeration generateCEnumeration(String enumerationName) {

        CEnumeration cEnumeration = (CEnumeration) _typeTable.getType("enum " + enumerationName);

        if (cEnumeration == null) {
            cEnumeration = _datatypesFactory.createCEnumeration();

            cEnumeration.setName(enumerationName);

            _typeTable.addType("enum " + enumerationName, cEnumeration);
        }

        return cEnumeration;
    }

    /**
     * 
     * Add a node representing a C structure to the object list
     * 
     * @param IASTClassSpecifier
     *            a structure node of the compilationUnit to be added to the EMF
     *            object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CEnumeration initializeCEnumeration(CEnumeration cEnumeration, IASTClassSpecifier enumToken, boolean topLevel) throws Exception {

        if (cEnumeration != null) {
            _typeTable.addSourceLocation(cEnumeration, new CSourceLocation(_fileName));

            Iterator enumerationIterator = enumToken.getDeclarations();

            while (enumerationIterator.hasNext()) {
                ASTEnumerator aSTEnumerator = (ASTEnumerator) enumerationIterator.next();

                CEnumerator cEnumerator = addEnumerator(aSTEnumerator);
                cEnumeration.getEnumerator().add(cEnumerator);
            }
        }

        return cEnumeration;
    }

    /**
     * 
     * Add a node representing a C structure to the object list
     * 
     * @param IASTClassSpecifier
     *            a structure node of the compilationUnit to be added to the EMF
     *            object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CEnumeration initializeCEnumeration(CEnumeration cEnumeration, ASTEnumerationSpecifier enumToken, boolean topLevel) throws Exception {

        if (cEnumeration != null) {
            _typeTable.addSourceLocation(cEnumeration, new CSourceLocation(_fileName));

            Iterator enumerationIterator = enumToken.getEnumerators();

            while (enumerationIterator.hasNext()) {
                ASTEnumerator aSTEnumerator = (ASTEnumerator) enumerationIterator.next();

                CEnumerator enumerator = addEnumerator(aSTEnumerator);
                cEnumeration.getEnumerator().add(enumerator);
            }
        }

        return cEnumeration;
    }

    /**
     * 
     * Add a node representing an enumerator to the object list
     * 
     * @param ASTTypedef
     *            a typedef node of the compilationUnit to be added to the EMF
     *            object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CEnumerator addEnumerator(ASTEnumerator enumeratorToken) throws Exception {

        CEnumerator cEnumerator = _factory.createCEnumerator();

        String enumeratorName = enumeratorToken.getName();
        cEnumerator.setName(enumeratorName);

        if (enumeratorToken.getInitialValue() != null) {
            IASTExpression iASTExpression = enumeratorToken.getInitialValue();
            String enumeratorValue = iASTExpression.toString();
            cEnumerator.setValue(new Integer(enumeratorValue));
        }

        return cEnumerator;
    }

    /**
     * 
     * Add a node representing a C typedef to the object list
     * 
     * @param ASTTypedef
     *            a typedef node of the compilationUnit to be added to the EMF
     *            object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CTypedef addTypedef(ASTTypedef typedefToken, boolean topLevel) throws Exception {

        String typedefName = typedefToken.getName();

        CTypedef cTypedef = generateCTypedef(typedefName);

        if (cTypedef != null && cTypedef.getDerives() == null) {

            cTypedef = initializeCTypedef(cTypedef, typedefToken, topLevel);

            _objectList.add(cTypedef);

            if (topLevel) {
                _topTypesList.add(cTypedef);
            }
        }
        return cTypedef;
    }

    /**
     * 
     * creates and returnes a CTypedef
     * 
     * @param String
     *            name of typedef
     *  
     */
    private CTypedef generateCTypedef(String typedefName) {

        CTypedef cTypedef = null;
        CDerivableType atype = _typeTable.getType(typedefName);

        if (atype == null) {
            cTypedef = null;
        } else if (atype instanceof CTypedef) {
            cTypedef = (CTypedef) atype;
        } else {
            return null;
        }

        if (cTypedef == null) {
            cTypedef = _factory.createCTypedef();
            cTypedef.setName(typedefName);
            _typeTable.addType(typedefName, cTypedef);
        }

        return cTypedef;
    }

    /**
     * 
     * Add a node representing a C structure to the object list
     * 
     * @param ASTTypedef
     *            a structure node of the compilationUnit to be added to the EMF
     *            object list
     * @param boolean
     *            true if passed node is a first level citizen
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CTypedef initializeCTypedef(CTypedef cTypedef, ASTTypedef typedefToken, boolean topLevel) throws Exception {

        if (cTypedef != null) {
            _typeTable.addSourceLocation(cTypedef, new CSourceLocation(_fileName));

            //
            // Add the deriving type
            //
            CDerivableType derivingType = null;

            Object typeSpecifierObject = typedefToken.getFinalTypeSpecifier();

            String name = getNameofComplexeType(typeSpecifierObject);

            try {
                derivingType = _typeParser.findOrCreateType(typedefToken, name, true);
                derivingType = initializeFieldType(derivingType, typeSpecifierObject);
            } catch (Exception e) {
                //
                // The first time thru ignore the errors, this might be a
                // forward reference.
                //
                if (!_firstPass) {
                    if (CPreferenceStore.isThrowExceptionAtParseError()) {
                        throw e;
                    } else {
                        CResource.writeLog(this, "addTypedef", e);
                    }
                }
            }

            if (derivingType != null) {
                cTypedef.setDerives(derivingType);
            } else if (!_firstPass) {
                CResource.writeLog(this, "addTypedef", CResource.getString("_ERROR_Type_Not_Found", "typedef", cTypedef.getName()));
            }
        }

        return cTypedef;
    }

    /**
     * 
     * Add a node representing a field of a structure/union to the object list
     * 
     * @param IASTField
     *            a field node of a structure/union
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CField addField(IASTField fieldToken) throws Exception {

        String fieldName = fieldToken.getName();
        Object lastElement = null;
        String localStructName = null;
        String localUnionName = null;

        CField cField = generateCField(fieldName);

        // Get the type of the field
        CDerivableType fieldType = null;

        try {
            IASTAbstractDeclaration iASTAbstractDeclaration = fieldToken.getAbstractDeclaration();
            Object fieldObject = iASTAbstractDeclaration.getTypeSpecifier();

            if (fieldObject instanceof ASTClassSpecifier || fieldObject instanceof ASTEnumerationSpecifier || fieldObject instanceof ASTElaboratedTypeSpecifier) {
                useFieldStack = true;

                addNode(fieldObject, false);

                if (!(addFieldStack.isEmpty())) {
                    lastElement = addFieldStack.lastElement();

                    if (lastElement != null && lastElement instanceof String) {
                        localStructName = (String) lastElement;
                        localUnionName = (String) lastElement;
                    }
                }
            }

            if ((localStructName == null || localStructName.length() <= 0) && currentStructureName != null) {
                localStructName = new String(currentStructureName);
            }

            if ((localUnionName == null || localUnionName.length() <= 0) && currentUnionName != null) {
                localUnionName = new String(currentUnionName);
            }

            fieldType = _typeParser.findOrCreateType(fieldToken, localStructName, localUnionName, false);

            fieldType = initializeFieldType(fieldType, fieldObject);

            if (localStructName != null) {
                currentStructureName = new String(localStructName);
            }

            if (localUnionName != null) {
                currentUnionName = new String(localUnionName);
            }

        } catch (Exception e) {

            if (CPreferenceStore.isThrowExceptionAtParseError()) {
                throw e;

            } else {
                CResource.writeLog(this, "addField", e);
            }
        }

        if (fieldType instanceof CClassifier) {
            cField.setType((CClassifier) fieldType);
        } else {
            CResource.writeLog(this, "addField", CResource.getString("_ERROR_Type_Not_Found", "field", fieldName));
        }

        useFieldStack = false;

        if (lastElement != null) {
            addFieldStack.remove(lastElement);
            lastElement = null;
        }

        return cField;
    }

    /**
     * 
     * Initializes the passed field type
     * 
     * @param CDerivableType
     *            field type to be initialized
     * @param Object
     *            contrains data for initialization
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    private CDerivableType initializeFieldType(CDerivableType fieldType, Object fieldObject) throws Exception {

        CStruct cStruct = null;
        CUnion cUnion = null;

        if (fieldObject instanceof ASTClassSpecifier) {

            if (fieldType instanceof CArray && ((CArray) fieldType).getDerives() != null) {

                if (((CArray) fieldType).getDerives() instanceof CStruct) {
                    cStruct = (CStruct) ((CArray) fieldType).getDerives();
                }

                if (((CArray) fieldType).getDerives() instanceof CUnion) {
                    cUnion = (CUnion) ((CArray) fieldType).getDerives();
                }

            } else if (fieldType instanceof CStruct) {
                cStruct = (CStruct) fieldType;

            } else if (fieldType instanceof CUnion) {
                cUnion = (CUnion) fieldType;
            }

            if (cStruct != null && (cStruct.getContains() != null)) {
                cStruct.getContains().clear();
                initializeCStruct(cStruct, (ASTClassSpecifier) fieldObject, false);

                _objectList.add(cStruct);
            }

            if (cUnion != null && (cUnion.getContains() != null)) {
                cUnion.getContains().clear();
                initializeCUnion(cUnion, (ASTClassSpecifier) fieldObject, false);

                _objectList.add(cUnion);
            }

        } else if (fieldObject instanceof ASTElaboratedTypeSpecifier) {

        }

        return fieldType;
    }

    /**
     * 
     * creates and returns a CField
     * 
     * @param String
     *            name of field
     *  
     */
    private CField generateCField(String fieldName) {

        CField cField = _factory.createCField();

        cField.setName(fieldName);

        return cField;
    }

    /**
     * 
     * getter method
     * 
     * @return java.util.ArrayList The list of C model objects created from the
     *         parsed information.
     *  
     */
    public ArrayList getObjectList() {
        return _objectList;
    }

    /**
     * 
     * getter method
     * 
     * @return String name of parsed C source file
     *  
     */
    public String getName() {
        return _fileName;
    }

    /**
     * 
     * getter method
     * 
     * @return java.util.ArrayList The list of included header files
     *  
     */
    public ArrayList getIncludeFiles() {
        return _includeFiles;
    }

    /**
     * 
     * getter method
     * 
     * @return CParseCallback parsed source
     *  
     */
    public CParseCallback getSourceParseCallback() {
        return _sourceParseCallback;
    }

    /**
     * Add a node representing a function parameter to the object list
     */
    private CParameter addParameter(CTypedElementParser atypedElement) throws Exception {

        CParameter aparameter = _factory.createCParameter();
        aparameter.setDirection(CDirectionKind.PARAMETER_LITERAL);
        aparameter.setName(atypedElement.getName());

        // Get the parameter type
        CDerivableType paramType = null;
        try {
            paramType = _typeParser.findOrCreateType(atypedElement.getType(), true);
        } catch (Exception e) {
            if (CPreferenceStore.isThrowExceptionAtParseError())
                throw e;
            else
                CResource.writeLog(this, "addParameter", e);
        }

        if (paramType instanceof CClassifier)
            aparameter.setType((CClassifier) paramType);
        else
            CResource.writeLog(this, "addParameter", CResource.getString("_ERROR_Type_Not_Found", "parameter", atypedElement.getName()));

        return aparameter;
    }

    /**
     * Add the return parameter of a function to the list
     */
    private CParameter addReturn(ArrayList returnList) throws Exception {

        if (returnList == null || returnList.size() == 0)
            return null;

        CParameter areturn = _factory.createCParameter();
        areturn.setDirection(CDirectionKind.RETURN_LITERAL);
        areturn.setName("");

        String returnTypeString = _typeParser.typeListToString(returnList);

        CDerivableType returnType = null;
        try {
            returnType = _typeParser.findOrCreateType(returnTypeString, true);
        } catch (Exception e) {
            if (CPreferenceStore.isThrowExceptionAtParseError())
                throw e;
            else
                CResource.writeLog(this, "addReturn", e);
        }

        if (returnType instanceof CClassifier)
            areturn.setType((CClassifier) returnType);
        else
            CResource.writeLog(this, "addReturn", CResource.getString("_ERROR_Type_Not_Found", "return parameter", returnTypeString));

        return areturn;
    }

    /**
     * Return a clone of the given CStruct object. This is used in the case of
     * nested structs.
     */
    private CStruct clonedStruct(CStruct astruct) {

        CStruct clonedStruct = _factory.createCStruct();
        clonedStruct.setName(astruct.getName());
        List fields = astruct.getContains();
        Iterator it = fields.iterator();
        while (it.hasNext()) {
            Object nextObject = it.next();
            if (nextObject instanceof CField) {
                CField afield = (CField) nextObject;
                CField clonedField = _factory.createCField();
                clonedField.setName(afield.getName());
                clonedField.setType(CQuery.typeOf(afield));
                if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
                    InstanceTDBase clonedInstance = CTypeDescriptorBuilder.clonedInstance(afield.getInstanceTDBase());
                    clonedField.setInstanceTDBase(clonedInstance);
                }
                clonedStruct.getContains().add(clonedField);
            } else if (nextObject instanceof CStruct) {
                CStruct nestedStruct = clonedStruct((CStruct) nextObject);
                clonedStruct.getContains().add(nestedStruct);
            } else if (nextObject instanceof CUnion) {
                CUnion nestedUnion = clonedUnion((CUnion) nextObject);
                clonedStruct.getContains().add(nestedUnion);
            }
        }
        return clonedStruct;
    }

    /**
     * Return a clone of the given CUnion object. This is used in the case of
     * nested unions.
     */
    private CUnion clonedUnion(CUnion aunion) {

        CUnion clonedUnion = _factory.createCUnion();
        clonedUnion.setName(aunion.getName());
        List fields = aunion.getContains();
        Iterator it = fields.iterator();
        while (it.hasNext()) {
            Object nextObject = it.next();
            if (nextObject instanceof CField) {
                CField afield = (CField) nextObject;
                CField clonedField = _factory.createCField();
                clonedField.setName(afield.getName());
                clonedField.setType(CQuery.typeOf(afield));
                if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
                    InstanceTDBase clonedInstance = CTypeDescriptorBuilder.clonedInstance(afield.getInstanceTDBase());
                    clonedField.setInstanceTDBase(clonedInstance);
                }
                clonedUnion.getContains().add(clonedField);
            } else if (nextObject instanceof CStruct) {
                CStruct nestedStruct = clonedStruct((CStruct) nextObject);
                clonedUnion.getContains().add(nestedStruct);
            } else if (nextObject instanceof CUnion) {
                CUnion nestedUnion = clonedUnion((CUnion) nextObject);
                clonedUnion.getContains().add(nestedUnion);
            }
        }
        return clonedUnion;
    }

    /**
     * This method gets the name of the passed complexe type.
     * 
     * @param IASTSimpleTypeSpecifier
     *            from which the name must be returned
     * 
     * @return String name of typeSpecifier
     */
    private String getNameofComplexeType(Object object) {

        String name = null;

        if (object instanceof IASTClassSpecifier) {
            name = getNameofComplexeType((IASTClassSpecifier) object);

        } else if (object instanceof ASTEnumerationSpecifier) {
            name = getNameofComplexeType((ASTEnumerationSpecifier) object);

        } else if (object instanceof ASTElaboratedTypeSpecifier) {
            name = getNameofComplexeType((ASTElaboratedTypeSpecifier) object);

        }

        return name;
    }

    /**
     * This method gets the name of the passed complexe type.
     * 
     * @param IASTClassSpecifier
     *            from which the name must be returned
     * 
     * @return String name of typeSpecifier
     */
    private String getNameofComplexeType(IASTClassSpecifier iASTClassSpecifier) {

        String name = null;

        if (iASTClassSpecifier != null) {
            ASTClassKind aSTClassKind = iASTClassSpecifier.getClassKind();

            if (aSTClassKind == ASTClassKind.STRUCT) {
                name = getNameofStructure(iASTClassSpecifier);

            } else if (aSTClassKind == ASTClassKind.UNION) {
                name = getNameofUnion(iASTClassSpecifier);

            } else if (aSTClassKind == ASTClassKind.ENUM) {
                name = getNameofEnum(iASTClassSpecifier);

            }
        }

        return name;
    }

    /**
     * This method gets the name of the passed complexe type.
     * 
     * @param ASTEnumerationSpecifier
     *            from which the name must be returned
     * 
     * @return String name of typeSpecifier
     */
    private String getNameofComplexeType(ASTEnumerationSpecifier aSTEnumerationSpecifier) {

        String name = null;

        if (aSTEnumerationSpecifier != null) {
            name = aSTEnumerationSpecifier.getName();
            name = getNameofEnum(name);
        }

        return name;
    }

    /**
     * This method gets the name of the passed complexe type.
     * 
     * @param ASTElaboratedTypeSpecifier
     *            from which the name must be returned
     * 
     * @return String name of typeSpecifier
     */
    private String getNameofComplexeType(ASTElaboratedTypeSpecifier aSTElaboratedTypeSpecifier) {

        String name = null;

        if (aSTElaboratedTypeSpecifier != null) {
            name = aSTElaboratedTypeSpecifier.getName();

            ASTClassKind aSTClassKind = aSTElaboratedTypeSpecifier.getClassKind();

            List list = aSTElaboratedTypeSpecifier.getReferences();

            if (aSTClassKind == ASTClassKind.STRUCT) {
                name = getNameofStructure(name);

            } else if (aSTClassKind == ASTClassKind.UNION) {
                name = getNameofUnion(name);

            } else if (aSTClassKind == ASTClassKind.ENUM) {
                name = getNameofEnum(name);
            }
        }

        return name;
    }

    /**
     * This method gets the name of the passed structure (IASTClassSpecifier).
     * If no name is defined, then a unique default name will be assigned.
     * 
     * @param IASTClassSpecifier
     *            from which the name must be returned
     * 
     * @return String name of typeSpecifier
     */
    private String getNameofStructure(IASTClassSpecifier classSpecifier) {

        String name = classSpecifier.getName();

        name = getNameofStructure(name);

        return name;
    }

    /**
     * If no name is defined, then a unique default name will be assigned.
     * 
     * @param IASTClassSpecifier
     *            from which the name must be returned
     * 
     * @return String name of typeSpecifier
     */
    private String getNameofStructure(String name) {

        if (name == null || (name != null && name.length() <= 0)) {
            name = STRUCTURE_TYPE_SPECIFIER + getStructureCounter();
            setStructureCounter(getStructureCounter() + 1);
        }

        return name;
    }

    /**
     * This method gets the name of the passed union (IASTClassSpecifier). If no
     * name is defined, then a unique default name will be assigned.
     * 
     * @param IASTClassSpecifier
     *            from which the name must be returned
     * 
     * @return String name of typeSpecifier
     */
    private String getNameofUnion(IASTClassSpecifier classSpecifier) {

        String name = classSpecifier.getName();

        name = getNameofUnion(name);

        return name;
    }

    /**
     * If no name is defined, then a unique default name will be assigned.
     * 
     * @param IASTClassSpecifier
     *            from which the name must be returned
     * 
     * @return String name of typeSpecifier
     */
    private String getNameofUnion(String name) {

        if (name == null || (name != null && name.length() <= 0)) {
            name = UNION_TYPE_SPECIFIER + getUnionCounter();
            setUnionCounter(getUnionCounter() + 1);
        }

        return name;
    }

    /**
     * This method gets the name of the passed enum (IASTClassSpecifier). If no
     * name is defined, then a unique default name will be assigned.
     * 
     * @param IASTClassSpecifier
     *            from which the name must be returned
     * 
     * @return String name of typeSpecifier
     */
    private String getNameofEnum(IASTClassSpecifier classSpecifier) {

        String name = classSpecifier.getName();

        name = getNameofEnum(name);

        return name;
    }

    /**
     * If no name is defined, then a unique default name will be assigned.
     * 
     * @param String
     *            from which the name must be returned
     * 
     * @return String name of typeSpecifier
     */
    private String getNameofEnum(String name) {

        if (name == null || (name != null && name.length() <= 0)) {
            name = ENUM_TYPE_SPECIFIER + getEnumCounter();
            setEnumCounter(getEnumCounter() + 1);
        }

        return name;
    }

    /**
     * @return int structureCounter
     */
    private int getStructureCounter() {
        return structureCounter;
    }

    /**
     * @param specifierCounter
     */
    private void setStructureCounter(int structureCounter) {
        this.structureCounter = structureCounter;
    }

    /**
     * @return int unionCounter
     */
    private int getUnionCounter() {
        return unionCounter;
    }

    /**
     * @param unionCounter
     */
    private void setUnionCounter(int unionCounter) {
        this.unionCounter = unionCounter;
    }

    /**
     * @return int enumCounter
     */
    private int getEnumCounter() {
        return enumCounter;
    }

    /**
     * @param enumCounter
     */
    private void setEnumCounter(int enumCounter) {
        this.enumCounter = enumCounter;
    }
}
