/*
 * Licensed Materials - Property of IBM
 *
 * com.ibm.etools.c
 *
 *  Copyright IBM Corporation 2004. All Rights Reserved. 
 *
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.etools.c.importer;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.ibm.etools.c.CClassifier;
import com.ibm.etools.c.CDerivableType;
import com.ibm.etools.c.CDerived;
import com.ibm.etools.c.CFactory;
import com.ibm.etools.c.CField;
import com.ibm.etools.c.CFunction;
import com.ibm.etools.c.CParameter;
import com.ibm.etools.c.CPointer;
import com.ibm.etools.c.CStruct;
import com.ibm.etools.c.CStructured;
import com.ibm.etools.c.CTypedef;
import com.ibm.etools.c.CUnion;
import com.ibm.etools.c.datatypes.CBitField;
import com.ibm.etools.typedescriptor.AddressTD;
import com.ibm.etools.typedescriptor.BaseTDType;
import com.ibm.etools.typedescriptor.InstanceTDBase;
import com.ibm.etools.typedescriptor.SimpleInstanceTD;

/**
 * The CUtility class provides utility methods for other C Importer classes.
 */
public class CUtility {

    public static final String copyright = "Licensed Material - Property of IBM\n com.ibm.etools.c\n(C) Copyright IBM Corp. 2004 - All Rights Reserved.\nUS Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp."; //$NON-NLS-1$

    private static boolean _platformCompilerInfoAdded = false;

    private CUtility() {
        super();
    }

    /**
     * Create a normalized type by removing unnecessary empty spaces.
     */
    public static String normalizedType(String str) {

        StringBuffer buffer = new StringBuffer(str.trim());
        int length = buffer.length();
        int index = 0;
        while (index < length) {
            char achar = buffer.charAt(index);
            if (Character.isWhitespace(achar)) {
                char nextChar = buffer.charAt(index + 1);
                if (Character.isWhitespace(nextChar) || nextChar == '*' || nextChar == '(' || nextChar == '[' || nextChar == ']' || buffer.charAt(index - 1) == '[') {
                    buffer.deleteCharAt(index);
                    length = length - 1;
                } else {
                    index = index + 1;
                }
            } else {
                index = index + 1;
            }
        }
        return buffer.toString();
    }

    /**
     * For a given string and a given position of bracket index, find the index
     * of the matching bracket. The search can be either forward or backward.
     * This is used for array processing.
     */
    public static int findMatchingBracket(String str, int indexOfBracket, boolean forward) {

        int index = indexOfBracket;
        if (forward) {
            index = index + 1;
            while (str.charAt(index) != ']')
                index = index + 1;
        } else {
            index = index - 1;
            while (str.charAt(index) != '[')
                index = index - 1;
        }

        return index;
    }

    /**
     * Return the full object list for the given object list. The CImporter
     * class calls this method before creating an Extent in order to meet the
     * MOF requirements on resources. If the Extent object list is incomplete,
     * the MOF driver will throw an exception if you try to save the resource
     * persistently.
     * 
     * Rules to generate the full object list: If object A is in the original
     * list and it has a reference to another object B, then object B should
     * also be added to the list if it is not already there.
     */
    public static ArrayList generateFullObjectList(ArrayList objectList) throws CException {

        _platformCompilerInfoAdded = false;
        ArrayList newList = new ArrayList();
        newList.addAll(objectList);

        Iterator iterator = objectList.iterator();
        while (iterator.hasNext()) {
            Object listElement = iterator.next();
            if (listElement instanceof CStruct) {
                CStruct astruct = (CStruct) listElement;
                List fieldList = astruct.getContains();
                addFieldsToList(fieldList, newList);
                if (CPreferenceStore.isCreateImplicitFieldsForTypes())
                    createImplicitFieldForType(astruct, newList, "_implict_instance_for_struct_" + astruct.getName());
            } else if (listElement instanceof CUnion) {
                CUnion aunion = (CUnion) listElement;
                List fieldList = aunion.getContains();
                addFieldsToList(fieldList, newList);
                if (CPreferenceStore.isCreateImplicitFieldsForTypes())
                    createImplicitFieldForType(aunion, newList, "_implict_instance_for_union_" + aunion.getName());
            } else if (listElement instanceof CTypedef) {
                CTypedef atypedef = (CTypedef) listElement;
                CDerivableType derivingType = atypedef.getDerives();
                if (derivingType != null)
                    addTypeToList(derivingType, newList);
                if (CPreferenceStore.isCreateImplicitFieldsForTypes())
                    createImplicitFieldForType(atypedef, newList, "_implict_instance_for_typedef_" + atypedef.getName());
            } else if (listElement instanceof CFunction) {
                CFunction afunction = (CFunction) listElement;
                List parameterList = afunction.getParameter();
                addParametersToList(parameterList, newList);
            } else if (listElement instanceof CField) {
                CField afield = (CField) listElement;
                CClassifier fieldType = CQuery.typeOf(afield);
                if (fieldType != null)
                    addTypeToList(fieldType, newList);
                InstanceTDBase instance = afield.getInstanceTDBase();
                if (instance != null)
                    addInstanceToList(instance, newList);
            } else {
                // Ignore other elements
            }
        }
        return newList;
    }

    /**
     * Create an implicit variable for a given type
     */
    private static void createImplicitFieldForType(CClassifier type, ArrayList objectList, String fieldName) throws CException {

        //CFactory factory = CImporter.instance().getFactory();
        CFactory factory = CFactory.eINSTANCE;
        CField afield = factory.createCField();
        afield.setName(fieldName);
        afield.setType(type);
        if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
            InstanceTDBase instance = CTypeDescriptorBuilder.createInstanceForVariable(afield);
            afield.setInstanceTDBase(instance);
            addInstanceToList(instance, objectList);
        }

        objectList.add(afield);
    }

    /**
     * Add the type of a field to the object list
     */
    private static void addFieldsToList(List fieldList, ArrayList objectList) {

        Iterator iterator = fieldList.iterator();
        while (iterator.hasNext()) {
            Object listElement = iterator.next();
            if (listElement instanceof CField) {
                CField afield = (CField) listElement;

                if (afield.getInstanceTDBase() != null) {
                    addInstanceToList(afield.getInstanceTDBase(), objectList);
                }

                CClassifier type = afield.getType();
                if (type != null) {
                    addTypeToList(type, objectList);
                }
            } else if (listElement instanceof CStructured) {
                CStructured astructured = (CStructured) listElement;
                addFieldsToList(astructured.getContains(), objectList);
            }

        }
    }

    /**
     * Add a type object to the object list if it is not already in the list If
     * the type is a derived type, then we also need to add the reference type
     * to the object list.
     */
    private static void addTypeToList(Object typeObject, ArrayList objectList) {

        if (typeObject == null)
            return;

        // Only do the add when the object is not already in the list
        if (!objectList.contains(typeObject)) {
            // If this is a derived type, also add the deriving type to the list
            if (typeObject instanceof CDerived) {
                if (typeObject instanceof CPointer && ((CPointer) typeObject).getDerives() instanceof CFunction) {
                    CFunction afunction = (CFunction) ((CPointer) typeObject).getDerives();
                    objectList.add(afunction);
                    List parameterList = afunction.getParameter();
                    addParametersToList(parameterList, objectList);
                }

                CDerived aderived = (CDerived) typeObject;
                CDerivableType derivingType = aderived.getDerives();
                if (derivingType != null) {
                    addTypeToList(derivingType, objectList);
                }
            } else if (typeObject instanceof CStructured) {
                CStructured astructured = (CStructured) typeObject;
                addFieldsToList(astructured.getContains(), objectList);
            } else if (typeObject instanceof CFunction) {
                CFunction afunction = (CFunction) typeObject;
                List parameterList = afunction.getParameter();
                addParametersToList(parameterList, objectList);
            } else if (typeObject instanceof CBitField) {
                CBitField abitfield = (CBitField) typeObject;
                addTypeToList(abitfield.getBaseType(), objectList);
            }

            objectList.add(typeObject);
        }
    }

    /**
     * Add the type of a parameter to the object list
     */
    private static void addParametersToList(List parameterList, ArrayList objectList) {

        Iterator iterator = parameterList.iterator();
        while (iterator.hasNext()) {
            Object listElement = iterator.next();
            if (listElement instanceof CParameter) {
                CParameter aparameter = (CParameter) listElement;

                if (aparameter.getInstanceTDBase() != null) {
                    addInstanceToList(aparameter.getInstanceTDBase(), objectList);
                }

                CClassifier type = aparameter.getType();
                if (type != null) {
                    addTypeToList(type, objectList);
                }
            }
        }
    }

    /**
     * Add a TypeDescriptor instance to the object list
     */
    private static void addInstanceToList(InstanceTDBase instance, ArrayList objectList) {

        if (!_platformCompilerInfoAdded) {
            objectList.add(instance.getPlatformInfo());
            _platformCompilerInfoAdded = true;
        }

        if (instance instanceof SimpleInstanceTD) {
            SimpleInstanceTD simp = (SimpleInstanceTD) instance;
            BaseTDType baseType = simp.getSharedType();
            if (baseType != null)
                addTDTypeToList(baseType, objectList);
        }

        objectList.add(instance);
    }

    /**
     * Add a TypeDescriptor type to the object list
     */
    private static void addTDTypeToList(BaseTDType tdType, ArrayList objectList) {

        if (!objectList.contains(tdType)) {
            if (tdType instanceof AddressTD) {
                BaseTDType referenceType = ((AddressTD) tdType).getReferenceType();
                if (referenceType != null)
                    addTDTypeToList(referenceType, objectList);
            }
            objectList.add(tdType);
        }

    }

}